/*********************************************************************
* Software License Agreement (BSD License)
* 
*  Copyright (c) 2008, Willow Garage, Inc.
*  All rights reserved.
* 
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
* 
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the Willow Garage nor the names of its
*     contributors may be used to endorse or promote products derived
*     from this software without specific prior written permission.
* 
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

/** \author Ioan Sucan, Sachin Chitta */

#include "planning_environment/monitors/planning_monitor.h"
#include "planning_environment/util/kinematic_state_constraint_evaluator.h"
#include <boost/scoped_ptr.hpp>

void planning_environment::PlanningMonitor::loadParams(void)
{
  nh_.param<double>("collision_map_safety_timeout", intervalCollisionMap_, 0.0);
  nh_.param<double>("joint_states_safety_timeout", intervalState_, 0.0);
  nh_.param<double>("tf_safety_timeout", intervalPose_, 0.0);
  nh_.param<int>("contacts_to_compute_for_allowable_contacts_test", num_contacts_allowable_contacts_test_, 10);
  nh_.param<int>("contacts_to_compute_for_display", num_contacts_for_display_, 1);
}

bool planning_environment::PlanningMonitor::isEnvironmentSafe(motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  if (!isMapUpdated(intervalCollisionMap_) && use_collision_map_)
  {
    ROS_WARN("Environment is not safe for motion: Collision map not updated in the last %f seconds", intervalCollisionMap_);
    error_code.val = error_code.SENSOR_INFO_STALE;
    return false;
  }
  
  if (!isJointStateUpdated(intervalState_))
  {
    ROS_WARN("Environment is not safe for motion: Robot state not updated in the last %f seconds", intervalState_);
    error_code.val = error_code.ROBOT_STATE_STALE;
    return false;
  }  

  if (includePose_)
    if (!isPoseUpdated(intervalPose_))
    {
      ROS_WARN("Environment is not safe for motion: Robot pose not updated in the last %f seconds", intervalPose_);
      error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
      return false;
    }
  error_code.val = error_code.SUCCESS;
  return true;
}

void planning_environment::PlanningMonitor::clearConstraints(void)
{
  path_constraints_.joint_constraints.clear();
  path_constraints_.position_constraints.clear();
  path_constraints_.orientation_constraints.clear();

  goal_constraints_.joint_constraints.clear();
  goal_constraints_.position_constraints.clear();
  goal_constraints_.orientation_constraints.clear();
}

void planning_environment::PlanningMonitor::setPathConstraints(const motion_planning_msgs::Constraints &constraints, 
                                                               motion_planning_msgs::ArmNavigationErrorCodes &error_code)
{
  path_constraints_ = constraints;
  transformConstraintsToFrame(path_constraints_, getFrameId(), error_code);
}

void planning_environment::PlanningMonitor::setGoalConstraints(const motion_planning_msgs::Constraints &constraints,
                                                               motion_planning_msgs::ArmNavigationErrorCodes &error_code)
{
  goal_constraints_ = constraints;
  transformConstraintsToFrame(goal_constraints_, getFrameId(), error_code);
}

bool planning_environment::PlanningMonitor::transformConstraintsToFrame(motion_planning_msgs::Constraints &constraints, 
                                                                        const std::string &target, 
                                                                        motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  geometry_msgs::PointStamped pos;
  geometry_msgs::PoseStamped tmp_pose;
  for (unsigned int i = 0; i < constraints.position_constraints.size() ; ++i)
  {
    bool ok = false;	
    unsigned int steps = 0;
    while (!tf_->canTransform(target, constraints.position_constraints[i].header.frame_id,ros::Time()) && steps < 10)
    {
      ros::Duration(0.01).sleep();
      steps++;
    }	
    try
    {
      pos.point = constraints.position_constraints[i].position;
      pos.header = constraints.position_constraints[i].header;
      pos.header.stamp = ros::Time();
      tf_->transformPoint(target,pos,pos);
      constraints.position_constraints[i].position = pos.point;

      geometry_msgs::QuaternionStamped tmp_quaternion;
      tmp_quaternion.quaternion = constraints.position_constraints[i].constraint_region_orientation;
      tmp_quaternion.header = constraints.position_constraints[i].header;
      tmp_quaternion.header.stamp = ros::Time();
      tf_->transformQuaternion(target,tmp_quaternion,tmp_quaternion);
      constraints.position_constraints[i].header.frame_id = tmp_quaternion.header.frame_id;
      constraints.position_constraints[i].header.stamp = ros::Time::now();
      ok = true;
    }
    catch(...)
    {
    }
	
    if (!ok)
    {
      ROS_ERROR("Unable to transform pose constraint on link '%s' to frame '%s'", constraints.position_constraints[i].link_name.c_str(), target.c_str());
      error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
      return false;
    }
  }

  for (unsigned int i = 0; i < constraints.orientation_constraints.size() ; ++i)
  {
    bool ok = false;	
    unsigned int steps = 0;
    while (!tf_->canTransform(target, constraints.orientation_constraints[i].header.frame_id, ros::Time()) && steps < 10)
    {
      ros::Duration(0.01).sleep();
      steps++;
    }
    try
    {
      geometry_msgs::QuaternionStamped orientation;
      orientation.header = constraints.orientation_constraints[i].header;
      orientation.header.stamp = ros::Time();
      orientation.quaternion = constraints.orientation_constraints[i].orientation;

      tf_->transformQuaternion(target, orientation, orientation);
      constraints.orientation_constraints[i].orientation = orientation.quaternion;
      constraints.orientation_constraints[i].header = orientation.header;
      ok = true;
    }
    catch(...)
    {
    }
	
    if (!ok)
    {
      ROS_ERROR("Unable to transform pose constraint on link '%s' to frame '%s'", constraints.orientation_constraints[i].link_name.c_str(), target.c_str());
      error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
      return false;
    }
  }    
  // if there are any floating or planar joints, transform them
  if (getKinematicModel()->getPlanarJoints().size() > 0 || getKinematicModel()->getFloatingJoints().size() > 0)
  {
    /*    for (unsigned int i = 0; i < constraints.joint_constraints.size() ; ++i)
          if (!transformJoint(constraints.joint_constraints[i].joint_name, 
          0, constraints.joint_constraints[i].position, 
          constraints.joint_constraints[i].header, target))
          {
          error_code.val = error_code.FRAME_TRANSFORM_FAILURE;        
          return false;
          }
    */
  }
  error_code.val = error_code.SUCCESS;
  return true;
}

bool planning_environment::PlanningMonitor::transformTrajectoryToFrame(trajectory_msgs::JointTrajectory &kp, 
                                                                       motion_planning_msgs::RobotState &robot_state, 
                                                                       const std::string &target, 
                                                                       motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{    
  // if there are no planar of floating transforms, there is nothing to do
  if (getKinematicModel()->getPlanarJoints().empty() && getKinematicModel()->getFloatingJoints().empty())
  {
    kp.header.frame_id = target;
    return true;
  }    
  roslib::Header updatedHeader = kp.header;
  updatedHeader.frame_id = target;
    
  // transform start state
  for (unsigned int i = 0 ; i < robot_state.joint_state.position.size() ; ++i)
    if (!transformJointToFrame(robot_state.joint_state.position[i], robot_state.joint_state.name[i], kp.header, target, error_code))
      return false;
    
    
  // transform the rest of the states

  // get the joints this path is for
  std::vector<planning_models::KinematicModel::Joint*> joints;
  joints.resize(kp.joint_names.size());
  for (unsigned int j = 0 ; j < joints.size() ; ++j)
  {
    joints[j] = getKinematicModel()->getJoint(kp.joint_names[j]);
    if (joints[j] == NULL)
    {
      ROS_ERROR("Unknown joint '%s' found on path", kp.joint_names[j].c_str());
      error_code.val = error_code.INVALID_TRAJECTORY;
      return false;
    }
  }
    
  // iterate through the states
  for (unsigned int i = 0 ; i < kp.points.size() ; ++i)
  {
    unsigned int u = 0;
    for (unsigned int j = 0 ; j < joints.size() ; ++j)
    {
      roslib::Header header = kp.header;
      if (!transformJoint(joints[j]->name, u, kp.points[i].positions[j], header, target, error_code))
      {
        error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
        return false;
      }
      updatedHeader = header;
      u += joints[j]->usedParams;
    }
  }
    
  kp.header = updatedHeader;
  return true;
}

bool planning_environment::PlanningMonitor::transformJointToFrame(double &value, 
                                                                  const std::string &joint_name, 
                                                                  roslib::Header &header, 
                                                                  const std::string &target,
                                                                  motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  return transformJoint(joint_name, 0, value, header, target, error_code);
}

bool planning_environment::PlanningMonitor::transformJoint(const std::string &name, 
                                                           unsigned int index, 
                                                           double &param, 
                                                           roslib::Header &header, 
                                                           const std::string& target, 
                                                           motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  // TODO
  // planar joints and floating joints may need to be transformed 
  planning_models::KinematicModel::Joint *joint = getKinematicModel()->getJoint(name);
  if (joint == NULL)
  {
    ROS_ERROR("Unknown joint '%s'", name.c_str());
    error_code.val = error_code.INVALID_TRAJECTORY;
    return false;
  }    
  header.frame_id = target;
  return true;
}

bool planning_environment::PlanningMonitor::isStateValidOnPath(const planning_models::KinematicState *state, 
                                                               bool verbose, 
							       motion_planning_msgs::ArmNavigationErrorCodes &error_code) const 
{
  ROS_DEBUG("Checking state for collisions and against path constraints");
  return isStateValid(state, COLLISION_TEST | PATH_CONSTRAINTS_TEST, verbose, error_code);    
}

bool planning_environment::PlanningMonitor::isStateValidAtGoal(const planning_models::KinematicState *state, 
                                                               bool verbose,
                                                               motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  ROS_DEBUG("Checking state for collisions and against path and goal constraints");
  return isStateValid(state, 
                      COLLISION_TEST | PATH_CONSTRAINTS_TEST | GOAL_CONSTRAINTS_TEST, 
                      verbose, 
                      error_code);    
}

bool planning_environment::PlanningMonitor::isStateValid(const planning_models::KinematicState *state, 
                                                         const int test, 
                                                         bool verbose, 
							 motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{   
  bool valid = true;
    
  getEnvironmentModel()->lock();
  getKinematicModel()->lock();
    
  bool vlevel = getEnvironmentModel()->getVerbose();
  getEnvironmentModel()->setVerbose(verbose);

  // figure out the poses of the robot model
  getKinematicModel()->computeTransforms(state->getParams());
  // update the collision space
  getEnvironmentModel()->updateRobotModel();

  // 
  // Check for collision
  //
  std::vector<collision_space::EnvironmentModel::Contact> contacts;
  if (test & COLLISION_TEST) {
   
    unsigned int numContacts = num_contacts_for_display_;
    if(!allowedContacts_.empty()) {
      numContacts = num_contacts_allowable_contacts_test_;
    }
    valid = !getEnvironmentModel()->getCollisionContacts(allowedContacts_, contacts, numContacts);    
    
    if (onCollisionContact_) {
      for (unsigned int i = 0 ; i < contacts.size() ; ++i)
        onCollisionContact_(contacts[i]);
    }
    if(!valid) 
    {
      error_code.val = error_code.COLLISION_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_WARN ("State is in collision.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }

  //
  // Check the joint limits
  //
  if (test & JOINT_LIMITS_TEST)
  {
    bool within_bounds = state->checkBoundsJoints();      
    if(!within_bounds)
    {
      error_code.val = error_code.JOINT_LIMITS_VIOLATED;
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }


  //
  // Check the path constraints
  //
  if (test & PATH_CONSTRAINTS_TEST)
  {	    
    KinematicConstraintEvaluatorSet constraint_evaluator;
    if (test & PATH_CONSTRAINTS_TEST)
    {
      constraint_evaluator.add(getKinematicModel(), path_constraints_.joint_constraints);
      constraint_evaluator.add(getKinematicModel(), path_constraints_.position_constraints);
      constraint_evaluator.add(getKinematicModel(), path_constraints_.orientation_constraints);
    }
    valid = constraint_evaluator.decide(state->getParams(), true);
    if(!valid) 
    {
      error_code.val = error_code.PATH_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_WARN("State violates path constraints.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }


  //
  // Check the goal constraints
  //
  if (test & GOAL_CONSTRAINTS_TEST)
  {	    
    KinematicConstraintEvaluatorSet goal_constraint_evaluator;
    ROS_DEBUG("Evaluating goal constraints: joint: %u, position: %u, orientation: %u",
              (unsigned int)goal_constraints_.joint_constraints.size(),
              (unsigned int)goal_constraints_.position_constraints.size(),
              (unsigned int)goal_constraints_.orientation_constraints.size());
    
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.joint_constraints);
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.position_constraints);
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.orientation_constraints);

    valid = goal_constraint_evaluator.decide(state->getParams(), true);
    if(!valid) 
    {
      error_code.val = error_code.GOAL_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_WARN("State violates goal constraints.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }    
  getEnvironmentModel()->setVerbose(vlevel);
  getKinematicModel()->unlock();
  getEnvironmentModel()->unlock();        
  return valid;    
}


bool planning_environment::PlanningMonitor::isStateValid(const motion_planning_msgs::RobotState &robot_state, 
                                                         const int test, 
                                                         bool verbose, 
							 motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{   
  //  planning_models::KinematicState sp(*(getRobotState()));


  bool valid = true;
    
  getEnvironmentModel()->lock();
  getKinematicModel()->lock();
    
  bool vlevel = getEnvironmentModel()->getVerbose();
  getEnvironmentModel()->setVerbose(verbose);

  planning_models::KinematicState *state = new planning_models::KinematicState(*getRobotState());
  state->setParamsJoints(robot_state.joint_state.position, robot_state.joint_state.name);

  // figure out the poses of the robot model
  getKinematicModel()->computeTransforms(state->getParams());
  // update the collision space
  getEnvironmentModel()->updateRobotModel();

  // 
  // Check for collisions
  //
  std::vector<collision_space::EnvironmentModel::Contact> contacts;
  if (test & COLLISION_TEST) {
    unsigned int numContacts = num_contacts_for_display_;
    if(!allowedContacts_.empty()) {
      numContacts = num_contacts_allowable_contacts_test_;
    }
    valid = !getEnvironmentModel()->getCollisionContacts(allowedContacts_, contacts, numContacts);    
    // If a callback exists for contact determination, call the callback for each of the found contacts                                                     
    if (onCollisionContact_) {
      for (unsigned int i = 0 ; i < contacts.size() ; ++i)
        onCollisionContact_(contacts[i]);
    }
    
    if(!valid) 
    {
      error_code.val = error_code.COLLISION_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_ERROR("State is in collision.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }

  //
  // Check the joint limits
  //
  if (test & JOINT_LIMITS_TEST)
  {
    bool within_bounds = state->checkBoundsJoints(robot_state.joint_state.name);      
    if(!within_bounds)
    {
      error_code.val = error_code.JOINT_LIMITS_VIOLATED;
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }


  //
  // Check the path constraints
  //
  if (test & PATH_CONSTRAINTS_TEST)
  {	    
    KinematicConstraintEvaluatorSet constraint_evaluator;
    if (test & PATH_CONSTRAINTS_TEST)
    {
      constraint_evaluator.add(getKinematicModel(), path_constraints_.joint_constraints);
      constraint_evaluator.add(getKinematicModel(), path_constraints_.position_constraints);
      constraint_evaluator.add(getKinematicModel(), path_constraints_.orientation_constraints);
    }
    valid = constraint_evaluator.decide(state->getParams(), true);
    if(!valid) 
    {
      error_code.val = error_code.PATH_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_WARN("State violates path constraints.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }


  //
  // Check the goal constraints
  //
  if (test & GOAL_CONSTRAINTS_TEST)
  {	    
    KinematicConstraintEvaluatorSet goal_constraint_evaluator;
    ROS_DEBUG("Evaluating goal constraints: joint: %u, position: %u, orientation: %u",
              (unsigned int)goal_constraints_.joint_constraints.size(),
              (unsigned int)goal_constraints_.position_constraints.size(),
              (unsigned int)goal_constraints_.orientation_constraints.size());
    
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.joint_constraints);
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.position_constraints);
    goal_constraint_evaluator.add(getKinematicModel(), goal_constraints_.orientation_constraints);

    valid = goal_constraint_evaluator.decide(state->getParams(), true);
    if(!valid) 
    {
      error_code.val = error_code.GOAL_CONSTRAINTS_VIOLATED;
      if(verbose)
        ROS_WARN("State violates goal constraints.\n");
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
  }    
  getEnvironmentModel()->setVerbose(vlevel);
  getKinematicModel()->unlock();
  getEnvironmentModel()->unlock();        
  return valid;    
}

int planning_environment::PlanningMonitor::closestStateOnTrajectory(const trajectory_msgs::JointTrajectory &trajectory, 
                                                                    motion_planning_msgs::RobotState &robot_state, 
                                                                    const planning_models::KinematicState *state,
                                                                    motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  return closestStateOnTrajectory(trajectory, robot_state, 0, trajectory.points.size() - 1, state, error_code);
}

int planning_environment::PlanningMonitor::closestStateOnTrajectory(const trajectory_msgs::JointTrajectory &trajectory, 
                                                                    motion_planning_msgs::RobotState &robot_state, 
                                                                    unsigned int start, 
                                                                    unsigned int end, 
                                                                    const planning_models::KinematicState *state,
                                                                    motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  if (end >= trajectory.points.size())
    end = trajectory.points.size() - 1;
  if (start > end)
  {
    ROS_ERROR("Invalid start %d and end %d specification",start,end);
    error_code.val = error_code.INVALID_INDEX;
    return -1;
  }

  if (trajectory.header.frame_id != getFrameId())
  {
    trajectory_msgs::JointTrajectory pathT = trajectory;
    if (transformTrajectoryToFrame(pathT, robot_state, getFrameId(), error_code))
      return closestStateOnTrajectoryAux(pathT, start, end, state, error_code);
    else
    {
      ROS_ERROR("Could not transform trajectory from %s to %s",trajectory.header.frame_id.c_str(),getFrameId().c_str());
      error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
      return -1;
    }
  }
  else
    return closestStateOnTrajectoryAux(trajectory, start, end, state, error_code);  
}

int planning_environment::PlanningMonitor::closestStateOnTrajectoryAux(const trajectory_msgs::JointTrajectory &trajectory, 
                                                                       unsigned int start, 
                                                                       unsigned int end, 
                                                                       const planning_models::KinematicState *state,
                                                                       motion_planning_msgs::ArmNavigationErrorCodes &error_code) const
{
  double dist = 0.0;
  int    pos  = -1;

  // get the joints this path is for
  std::vector<planning_models::KinematicModel::Joint*> joints(trajectory.joint_names.size());
  for (unsigned int j = 0 ; j < joints.size() ; ++j)
  {
    joints[j] = getKinematicModel()->getJoint(trajectory.joint_names[j]);
    if (joints[j] == NULL)
    {
      ROS_ERROR("Unknown joint '%s' found on path", trajectory.joint_names[j].c_str());
      error_code.val = error_code.INVALID_TRAJECTORY;
      return -1;
    }
  }
    
  for (unsigned int i = start ; i <= end ; ++i)
  {
    double       d = 0.0;
    for (unsigned int j = 0 ; j < joints.size() ; ++j)
    {
	    
      double current_joint_position = (state->getParamsJoint(joints[j]->name))[0];	    
      double diff = fabs(trajectory.points[i].positions[j] - current_joint_position);
      d += diff * diff;
    }
	
    if (pos < 0 || d < dist)
    {
      pos = i;
      dist = d;
    }
  }    
  return pos;
}



bool planning_environment::PlanningMonitor::isTrajectoryValid(const trajectory_msgs::JointTrajectory &trajectory, 
                                                              motion_planning_msgs::RobotState &robot_state, 
                                                              const int test, 
                                                              bool verbose, 
                                                              motion_planning_msgs::ArmNavigationErrorCodes &error_code, 
                                                              std::vector<motion_planning_msgs::ArmNavigationErrorCodes> &trajectory_error_codes) const
{
  return isTrajectoryValid(trajectory, 
                           robot_state, 
                           0, 
                           trajectory.points.size() - 1, 
                           test, 
                           verbose,
                           error_code,trajectory_error_codes);
}

bool planning_environment::PlanningMonitor::isTrajectoryValid(const trajectory_msgs::JointTrajectory &trajectory, 
                                                              motion_planning_msgs::RobotState &robot_state, 
                                                              unsigned int start, 
                                                              unsigned int end, 
                                                              const int test, 
                                                              bool verbose,
                                                              motion_planning_msgs::ArmNavigationErrorCodes &error_code,
                                                              std::vector<motion_planning_msgs::ArmNavigationErrorCodes> &trajectory_error_codes) const
{
  if (end >= trajectory.points.size())
    end = trajectory.points.size() - 1;
  if (start > end)
  {
    ROS_ERROR("Invalid trajectory: start: %d, end: %d",start,end);
    error_code.val = error_code.INVALID_INDEX;
    return true;
  }
  if (trajectory.header.frame_id != getFrameId())
  {
    trajectory_msgs::JointTrajectory pathT = trajectory;
    if (transformTrajectoryToFrame(pathT, robot_state, getFrameId(), error_code))
      return isTrajectoryValidAux(pathT, robot_state, start, end, test, verbose, error_code, trajectory_error_codes);
    else
    { 
      ROS_WARN("Could not transform trajectory from frame: %s to frame: %s",pathT.header.frame_id.c_str(),getFrameId().c_str());
      error_code.val = error_code.FRAME_TRANSFORM_FAILURE;
      return false;
    }
  }
  else
    return isTrajectoryValidAux(trajectory, robot_state, start, end, test, verbose, error_code, trajectory_error_codes); 
}

bool planning_environment::PlanningMonitor::getRobotStateMsg(motion_planning_msgs::RobotState &robot_state)
{
  robot_state.joint_state.position.clear();
  robot_state.joint_state.name.clear();
  planning_models::KinematicState sp(*(getRobotState()));
  // boost::scoped_ptr<planning_models::KinematicState> sp(new planning_models::KinematicState(getKinematicModel()));
  // fill in robot state with current one
  std::vector<const planning_models::KinematicModel::Joint*> joints;
  getKinematicModel()->getJoints(joints);
	    
  robot_state.joint_state.name.resize(joints.size());
  robot_state.joint_state.position.resize(joints.size());
  robot_state.joint_state.header.frame_id = getFrameId();
  robot_state.joint_state.header.stamp = lastJointStateUpdate();

  for (unsigned int i = 0 ; i < joints.size() ; ++i)
  {
    robot_state.joint_state.name[i] = joints[i]->name;
    std::vector<double> tmp;
    sp.copyParamsJoint(tmp, joints[i]->name);
    if(!tmp.empty())
      robot_state.joint_state.position[i] = tmp[0];
  }	    
  return true;
}

bool planning_environment::PlanningMonitor::setRobotState(const motion_planning_msgs::RobotState &robot_state)
{
  //  boost::scoped_ptr<planning_models::KinematicState> sp(new planning_models::KinematicState(getKinematicModel()));
  planning_models::KinematicState sp(*(getRobotState()));

  std::vector<double> tmp;
  tmp.resize(1);

  for (unsigned int i = 0 ; i < robot_state.joint_state.position.size() ; ++i)
  {
    tmp[0] = robot_state.joint_state.position[i];
    sp.setParamsJoint(tmp, robot_state.joint_state.name[i]);
  }

  if (!sp.seenAll())// check whether the robot state specification contained all the joints we know about
  {
    ROS_WARN("Incomplete start state specification in path");
    return false;
  }
  return true;
}

bool planning_environment::PlanningMonitor::setJointStateAndComputeTransforms(const sensor_msgs::JointState &joint_state)
{
  //  boost::scoped_ptr<planning_models::KinematicState> sp(new planning_models::KinematicState(getKinematicModel()));
  planning_models::KinematicState sp(*(getRobotState()));

  std::vector<double> tmp;
  tmp.resize(1);
  for (unsigned int i = 0 ; i < joint_state.position.size() ; ++i)
  {
    tmp[0] = joint_state.position[i];
    sp.setParamsJoint(tmp, joint_state.name[i]);
  }
  getKinematicModel()->computeTransforms(sp.getParams());
  return true;
}

bool planning_environment::PlanningMonitor::setJointState(const sensor_msgs::JointState &joint_state)
{
  planning_models::KinematicState sp(*(getRobotState()));

  std::vector<double> tmp;
  tmp.resize(1);
  for (unsigned int i = 0 ; i < joint_state.position.size() ; ++i)
  {
    tmp[0] = joint_state.position[i];
    sp.setParamsJoint(tmp, joint_state.name[i]);
  }
  getKinematicModel()->computeTransforms(sp.getParams());
  return true;
}

bool planning_environment::PlanningMonitor::isTrajectoryValidAux(const trajectory_msgs::JointTrajectory &trajectory, 
                                                                 motion_planning_msgs::RobotState &robot_state, 
                                                                 unsigned int start, 
                                                                 unsigned int end, 
                                                                 const int test, 
                                                                 bool verbose,
                                                                 motion_planning_msgs::ArmNavigationErrorCodes &error_code,
                                                                 std::vector<motion_planning_msgs::ArmNavigationErrorCodes> &trajectory_error_codes) const
{    
  boost::scoped_ptr<planning_models::KinematicState> sp(new planning_models::KinematicState(getKinematicModel()));    
  if(trajectory.points.empty())
  {
    ROS_ERROR("Trajectory was empty");
    error_code.val = error_code.INVALID_TRAJECTORY;
    return false;
  }
  ROS_DEBUG("Checking trajectory in planning monitor");
  std::vector<double> tmp;
  tmp.resize(1);

  for (unsigned int i = 0 ; i < robot_state.joint_state.position.size() ; ++i)
  {
    tmp[0] = robot_state.joint_state.position[i];
    sp->setParamsJoint(tmp, robot_state.joint_state.name[i]);
  }

  if (!sp->seenAll())// check whether the robot state specification contained all the joints we know about
  {
    ROS_WARN("Incomplete start state specification in trajectory");
    error_code.val = error_code.INCOMPLETE_ROBOT_STATE;
    return false;
  }
    
  KinematicConstraintEvaluatorSet constraint_evaluator;
  if (test & PATH_CONSTRAINTS_TEST)
  {
    constraint_evaluator.add(getKinematicModel(), path_constraints_.joint_constraints);
    constraint_evaluator.add(getKinematicModel(), path_constraints_.position_constraints);
    constraint_evaluator.add(getKinematicModel(), path_constraints_.orientation_constraints);
  }
    
  getEnvironmentModel()->lock();
  getKinematicModel()->lock();


  bool vlevel = getEnvironmentModel()->getVerbose();
  getEnvironmentModel()->setVerbose(verbose);
       
  // figure out the poses of the robot model
  getKinematicModel()->computeTransforms(sp->getParams());
  // update the collision space
  getEnvironmentModel()->updateRobotModel();
    
  bool valid = true;
    
  // get the joints this trajectory is for
  std::vector<planning_models::KinematicModel::Joint*> joints(trajectory.joint_names.size());
  unsigned int sdim = 0;
  for (unsigned int j = 0 ; j < joints.size() ; ++j)
  {
    joints[j] = getKinematicModel()->getJoint(trajectory.joint_names[j]);
    if (joints[j] == NULL)
    {
      ROS_ERROR("Unknown joint '%s' found on path", trajectory.joint_names[j].c_str());
      error_code.val = error_code.INVALID_TRAJECTORY;
      getEnvironmentModel()->unlock();
      getKinematicModel()->unlock();
      return false;
    }
    sdim += joints[j]->usedParams;
  }

  //unsigned int remainingContacts = maxCollisionContacts_;    
  // check every state
  trajectory_error_codes.resize(trajectory.points.size());
  bool check_all = (test & CHECK_FULL_TRAJECTORY);
  bool valid_all = true;
  for (unsigned int i = start ; i <= end ; ++i)
  {
    if (trajectory.points[i].positions.size() != joints.size())
    {
      ROS_ERROR("Incorrect state specification on trajectory");
      error_code.val = error_code.INVALID_TRAJECTORY;      
      trajectory_error_codes[i].val = error_code.INVALID_TRAJECTORY;
      valid = false;
      valid_all = valid && valid_all;
      if(check_all)
        continue;
      else
        break;
    }
    sp->setParamsJoints(trajectory.points[i].positions, trajectory.joint_names);		    
    getKinematicModel()->computeTransforms(sp->getParams());
    getEnvironmentModel()->updateRobotModel();

    //
    // check the joint limits
    //
    if (test & JOINT_LIMITS_TEST)
    {
      valid = sp->checkBoundsJoints(trajectory.joint_names);
      valid_all = valid && valid_all;
      if(!valid)
      {
        error_code.val = error_code.JOINT_LIMITS_VIOLATED;
        trajectory_error_codes[i].val = error_code.JOINT_LIMITS_VIOLATED;
        ROS_ERROR("Joint limits violated");
        if(check_all)
          continue;
        else
          break;
      }
    }

    //
    // check for collisions
    //
    std::vector<collision_space::EnvironmentModel::Contact> contacts;
    if (test & COLLISION_TEST) 
    {
      unsigned int numContacts = num_contacts_for_display_;
      if(!allowedContacts_.empty()) {
        numContacts = num_contacts_allowable_contacts_test_;
      }
      valid = !getEnvironmentModel()->getCollisionContacts(allowedContacts_, contacts, numContacts);    
      
      if (onCollisionContact_) {
        for (unsigned int i = 0 ; i < contacts.size() ; ++i)
          onCollisionContact_(contacts[i]);
      }
      
      valid_all = valid && valid_all;
    }
    if(!valid)
    {
      //        if(!this->broadcastCollisions())
      //          ROS_WARN("Could not broadcast solutions");
      ROS_DEBUG("Found a collision for trajectory index: %d",i);
      error_code.val = error_code.COLLISION_CONSTRAINTS_VIOLATED;
      trajectory_error_codes[i].val = error_code.COLLISION_CONSTRAINTS_VIOLATED;
      if(check_all)
        continue;
      else
        break;
    }
    //
    // check against the path constraints
    //
    if (test & PATH_CONSTRAINTS_TEST)
    {
      valid = constraint_evaluator.decide(sp->getParams());
      valid_all = valid && valid_all;
      if (!valid)
      {
        if(verbose)
          ROS_INFO("isTrajectoryValid: State %d does not satisfy path constraints", i);
        error_code.val = error_code.PATH_CONSTRAINTS_VIOLATED;
        trajectory_error_codes[i].val = error_code.COLLISION_CONSTRAINTS_VIOLATED;
        if(check_all)
          continue;
        else
          break;
      }
    }	    
  }

  //
  // check against the goal constraints
  //
  if (valid_all && (test & GOAL_CONSTRAINTS_TEST))
  {
    constraint_evaluator.add(getKinematicModel(), goal_constraints_.joint_constraints);
    constraint_evaluator.add(getKinematicModel(), goal_constraints_.position_constraints);
    constraint_evaluator.add(getKinematicModel(), goal_constraints_.orientation_constraints);
    valid = constraint_evaluator.decide(sp->getParams());
    valid_all = valid && valid_all;
    if (!valid)
    {
      ROS_WARN("isTrajectoryValid: Goal state does not satisfy goal constraints");
      error_code.val = error_code.GOAL_CONSTRAINTS_VIOLATED;
    }
  }

  getEnvironmentModel()->setVerbose(vlevel);
  getKinematicModel()->unlock();
  getEnvironmentModel()->unlock();    
  return valid_all;
}

bool planning_environment::PlanningMonitor::broadcastCollisions() {
  
  unsigned int numContacts = num_contacts_for_display_;
  if(!allowedContacts_.empty()) {
    numContacts = num_contacts_allowable_contacts_test_;
  }
  //this just broadcasts collisions for the most recent collision space calls
  if(onCollisionContact_) {
    std::vector<collision_space::EnvironmentModel::Contact> contacts;
    getEnvironmentModel()->getCollisionContacts(allowedContacts_, contacts, numContacts); 
    ROS_DEBUG("Callback defined with %u contacts", (unsigned int) contacts.size());
    for (unsigned int i = 0 ; i < contacts.size() ; ++i) {
      onCollisionContact_(contacts[i]);
    } 
  } 
  else 
  {
    return false;
  }
  return true;
}

void planning_environment::PlanningMonitor::setAllowedContacts(const std::vector<collision_space::EnvironmentModel::AllowedContact> &allowedContacts)
{
  allowedContacts_ = allowedContacts;
}

void planning_environment::PlanningMonitor::setAllowedContacts(const std::vector<motion_planning_msgs::AllowedContactSpecification> &allowed_contacts)
{  
  allowedContacts_.clear();
  for (unsigned int i = 0 ; i < allowed_contacts.size() ; ++i)
  {
    collision_space::EnvironmentModel::AllowedContact ac;
    if (computeAllowedContact(allowed_contacts[i], ac))
      allowedContacts_.push_back(ac);
  }
}

const std::vector<collision_space::EnvironmentModel::AllowedContact>& planning_environment::PlanningMonitor::getAllowedContacts(void) const
{
  return allowedContacts_;
}

void planning_environment::PlanningMonitor::printAllowedContacts(std::ostream &out)
{
  out << allowedContacts_.size() << " allowed contacts" << std::endl;
  for (unsigned int i = 0 ; i < allowedContacts_.size() ; ++i)
  {
    out << "  - allowing contacts up to depth " << allowedContacts_[i].depth << " between links: [";
    for (unsigned int j = 0 ; j < allowedContacts_[i].links.size() ; ++j)
      out << allowedContacts_[i].links[j];
    out << "] and bound " << allowedContacts_[i].bound.get() << std::endl;
  }
}

void planning_environment::PlanningMonitor::printConstraints(std::ostream &out)
{
  out << "Path constraints:" << std::endl;

  KinematicConstraintEvaluatorSet constraint_evaluator;
  constraint_evaluator.add(getKinematicModel(), path_constraints_.joint_constraints);
  constraint_evaluator.add(getKinematicModel(), path_constraints_.position_constraints);
  constraint_evaluator.add(getKinematicModel(), path_constraints_.orientation_constraints);
  constraint_evaluator.print(out);

  out << "Goal constraints:" << std::endl;
  constraint_evaluator.clear();
  constraint_evaluator.add(getKinematicModel(), goal_constraints_.joint_constraints);
  constraint_evaluator.add(getKinematicModel(), goal_constraints_.position_constraints);
  constraint_evaluator.add(getKinematicModel(), goal_constraints_.orientation_constraints);
  constraint_evaluator.print(out);
}

void planning_environment::PlanningMonitor::clearAllowedContacts(void)
{
  allowedContacts_.clear();
}

void planning_environment::PlanningMonitor::setCollisionCheck(const std::string link_name, bool state)
{
  getEnvironmentModel()->setCollisionCheck(link_name,state);
}

void planning_environment::PlanningMonitor::setCollisionCheckAll(bool state)
{
  getEnvironmentModel()->setCollisionCheckAll(state);
}

void planning_environment::PlanningMonitor::setCollisionCheckLinks(const std::vector<std::string> &link_names, bool state)
{
  getEnvironmentModel()->setCollisionCheckLinks(link_names,state);
}

void planning_environment::PlanningMonitor::setCollisionCheckOnlyLinks(const std::vector<std::string> &link_names, bool state)
{
  getEnvironmentModel()->setCollisionCheckOnlyLinks(link_names,state);
}

void planning_environment::PlanningMonitor::getChildLinks(const std::vector<std::string> &joints, 
							  std::vector<std::string> &link_names)
{
  std::set<std::string> links;
  for(unsigned int i=0; i < joints.size(); i++)
  {
    planning_models::KinematicModel::Joint *joint = getKinematicModel()->getJoint(joints[i]);
    if(joint)
      {
      if(joint->after)
	{
	  std::vector<std::string> child_links;
	  links.insert(joint->after->name);
	  joint->after->getAllChildLinksRecursive(child_links);
	  for(unsigned int j=0; j < child_links.size(); j++)
	    links.insert(child_links[j]);
	}    
      }	
  }
  for(std::set<std::string>::iterator set_iterator = links.begin(); set_iterator!= links.end(); set_iterator++)
  {
    link_names.push_back(*set_iterator);
  }
  return;
}

void planning_environment::PlanningMonitor::getOrderedCollisionOperationsForOnlyCollideLinks(const std::vector<std::string> &collision_check_links, 
                                                                                             const motion_planning_msgs::OrderedCollisionOperations &requested_collision_operations,
                                                                                             motion_planning_msgs::OrderedCollisionOperations &result_collision_operations)
{
  motion_planning_msgs::OrderedCollisionOperations result;
  motion_planning_msgs::CollisionOperation op;
  std::vector<motion_planning_msgs::CollisionOperation> self_collisions;

  //this disables everything and everything
  op.object1 = op.COLLISION_SET_ALL;
  op.object2 = op.COLLISION_SET_ALL;
  op.operation = op.operation = motion_planning_msgs::CollisionOperation::DISABLE;;
  result.collision_operations.push_back(op);
  
  //this enables collision_check_links with everything
  for(std::vector<std::string>::const_iterator it = collision_check_links.begin();
      it != collision_check_links.end();
      it++) 
  {
    op.object1 = (*it);
    op.object2 = op.COLLISION_SET_ALL;
    op.operation = motion_planning_msgs::CollisionOperation::ENABLE;
    result.collision_operations.push_back(op);
  } 

  std::vector<std::vector<bool> > current_allowed;
  std::map<std::string, unsigned int> vec_indices;
  //this disables collision_check_links with things they should not collide with
  getEnvironmentModel()->getDefaultAllowedCollisionMatrix(current_allowed, vec_indices);
  for(std::vector<std::string>::const_iterator it = collision_check_links.begin();
      it != collision_check_links.end();it++)
  {
    std::map<std::string, unsigned int>::iterator map_it = vec_indices.find(*it);
    if(map_it == vec_indices.end())
      continue;
    unsigned int index = map_it->second;
    for(std::map<std::string, unsigned int>::iterator index_it = vec_indices.begin(); index_it != vec_indices.end(); index_it++)
    {
      if(current_allowed[index][index_it->second])
      {	
        op.object1 = (*it);
        op.object2 = index_it->first;
        op.operation = motion_planning_msgs::CollisionOperation::DISABLE;
        result.collision_operations.push_back(op);    
      }
    }
  }  

  //this adds extra requested collision operations
  for(std::vector<motion_planning_msgs::CollisionOperation>::const_iterator it = requested_collision_operations.collision_operations.begin();
      it != requested_collision_operations.collision_operations.end();it++)
  {
    result.collision_operations.push_back(*it);
  }  

  result_collision_operations.collision_operations = result.collision_operations;

}
